/* an MPI/C program that computes a correlation function between 
*  two sets of values. 
*  Input: None.
*  Output: a correlation function between two sets of values which the 
*  two loops depend on.  
*   

*   Author : Vidyut Tarkas
*            vtarkas@mailbox.syr.edu
*/

/***********************************************************

  Size of 2 N-bodies is assumed to be 1000

  **********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

#define N 10    /*  The number of elements in the set of measurements  */

/*  The function that is used to compute the correlation  */
double f(double a, double b)
{
        return (a*a + b*b);
}

void main(int argc, char* argv[])

{ 
        int n;                   /* specifies the number of values per
						processor */
	int myRank;                 /* Rank of this process */
        double x[N],y[N];           /* Arrays containing the two sets of
						measurements */
        double local_correlation[N];/* Correlation array for this
						processor */
        int destProcessRank;        /* Rank of process to which data is to
						be sent */
        int sourceProcessRank;      /* Rank of process from which data is
						to be received */

	double yArray[N];            /* Temporary array used to
						receive y's from other processes */
        double sum;                 /* To sum up the elements of the
						correlation array for each processor */
        double correlation;         /* To hold the sum of all the
					correlation functions calculated by each process */
        int p;                      
        int i,j,k;                  

	MPI_Status status;
                
        /* Initialise MPI */ 
        MPI_Init(&argc, &argv);

        /* Find the number of processors being used */
        MPI_Comm_size(MPI_COMM_WORLD, &p);

        /* Get the rank of this process*/
	MPI_Comm_rank(MPI_COMM_WORLD, &myRank);

	/* Each processor handles N/p values */      
        n = N/p;
        
        /* Initialize arrays */
        for (i=0; i<N;i++) 
	{
            local_correlation[i] = 0;
            yArray[i]=0;
	    y[i] = myRank*n + i;
	    x[i] = myRank*n + i;
	}
        sum = 0; 
	correlation=0;   
	for (k=0; k<p; k++) 
	{	
	    for ( j=0; j<n; j++)
		for ( i=0; i<n; i++)             
		  local_correlation[i] += f(x[i],y[j]);
                               
	    destProcessRank = (myRank+1);
	    if (destProcessRank >= p)
		 destProcessRank = 0;
	    sourceProcessRank = myRank-1;
	    if (sourceProcessRank < 0)
		 sourceProcessRank = p-1;
	    MPI_Sendrecv(&y,n, MPI_DOUBLE, destProcessRank ,0,
                  &yArray,n, MPI_DOUBLE, sourceProcessRank ,0, MPI_COMM_WORLD, &status); 
                  
	    for(i = 0; i < n; i++)	
		 y[i] = yArray[i];              
	}
     
        /*   summing up the elements of the correlation array  by each process  */
        for(i = 0; i < n; i++)
	   sum += local_correlation[i];	

        /*    a global reduction operation carried out to sum up the
               correlation functions obtained from all the processors  */
	MPI_Reduce(&sum, &correlation, 1, MPI_DOUBLE, MPI_SUM, 0,MPI_COMM_WORLD);

	if (myRank == 0)
	    printf("\nThe correlation function is %f\n\n",correlation); 

	MPI_Finalize();
}
